#ifndef VM86_H
#define VM86_H

#include <asm/vm86.h>
#define CF 0x001 /* C */
#define PF 0x004 /* P */
#define AF 0x010 /* A */
#define ZF 0x040 /* Z */
#define SF 0x080 /* S */
#define TF 0x100 /* T */
#define IF 0x200 /* I */
#define DF 0x400 /* D */
#define OF 0x800 /* O */

#endif /* VM86_H */
